from .utils import binary_feedback_handler, correction_feedback_handler, print_feedback, feedback, save_feedback_to_csv
from llmlink.model import BaseModel
import gradio as gr


class App:
    """
    An App class that wraps a model to be used in a Gradio chat interface

    Parameters
    ----------
    model : chainlink Model or custom class that extends the BaseModel class
        The model to be used with the application
    feedback : bool (default False)
        Whether to collect feedback within the app
    theme : str or None
        The gradio theme to use
    feedback_callback : function (default print_feedback)
        Function to use when feedback is submitted
    """

    def __init__(
            self,
            model,
            feedback=False,
            theme=None,
            feedback_callback=print_feedback
    ):
        self.model = model
        self.feedback = feedback
        self.theme = theme
        self.feedback_callback = feedback_callback

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        #    if not isinstance(value, BaseModel):
        #        raise TypeError('model must inherit from BaseModel class')
        self._model = value

    @property
    def feedback(self):
        return self._feedback

    @feedback.setter
    def feedback(self, value):
        if not isinstance(value, bool):
            raise TypeError('feedback must be bool')
        self._feedback = value

    @property
    def theme(self):
        return self._theme

    @theme.setter
    def theme(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError('theme must be str')
        self._theme = value

    def deploy(self, *args, **kwargs):
        """
        Deploy the application

        Parameters
        ----------
        *args : various arguments
            Arguments to send to gradio.Blocks().launch()
        **kwargs : various keyword arguments
            Keyword arguments to send to gradio.Blocks().launch()

        Returns
        -------
        deploy_results
            The results of running the gradio deploy function
        """
        if not self.feedback:
            with gr.Blocks(theme=self.theme) as demo:
                chatbot = gr.Chatbot(label='ChainLink App')
                msg = gr.Textbox(
                    lines=1,
                    max_lines=20,
                    placeholder='Input your text here.'
                )
                clear = gr.ClearButton([msg, chatbot])
                gr.Blocks()

                def respond(message, chat_history):
                    try:
                        bot_message = self.model.run(message)
                        chat_history.append((message, bot_message))
                        return "", chat_history
                    except Exception as e:
                        print(e)
                        chat_history.append(
                            (message, '**There was an error communicating with the bot**'))
                        return "", chat_history

                msg.submit(respond, [msg, chatbot], [msg, chatbot])

            return demo.launch(*args, **kwargs)

        else:
            with gr.Blocks(theme=self.theme) as demo:
                chatbot = gr.Chatbot(label='ChainLink App', elem_id='chatbot')

                with gr.Row():
                    with gr.Column(scale=100):
                        msg = gr.Textbox(
                            lines=1,
                            max_lines=20,
                            show_label=False,
                            placeholder='Input your text here.',
                            elem_id='user_msg_input'
                        )
                with gr.Row():
                    clear = gr.ClearButton([msg, chatbot])

                with gr.Row():
                    with gr.Column():
                        binary = gr.Radio(
                            ['Good Response', 'Bad Response'],
                            label='Was this a good response?',
                            value=None,
                            elem_id='binary_feedback_selector'
                        )

                with gr.Row():
                    with gr.Column(scale=100):
                        correction = gr.Textbox(
                            show_label=True,
                            label='Correction (optional)',
                            placeholder='Enter your correction here',
                            elem_id='user_correction_input'
                        )

                with gr.Row():
                    with gr.Column(scale=100):
                        submit_button = gr.Button(
                            value='Submit Feedback', elem_id='submit_feedback_button')

                def respond(message, chat_history):
                    try:
                        bot_message = self.model.run(message)
                        chat_history.append((message, bot_message))
                        feedback['user_msg'] = message
                        feedback['bot_msg'] = bot_message
                        return "", chat_history
                    except Exception as e:
                        print(e)
                        chat_history.append(
                            (message, '**There was an error communicating with the bot**'))
                        return "", chat_history

                msg.submit(respond, [msg, chatbot], [msg, chatbot])
                binary.change(fn=lambda choice: binary_feedback_handler(
                    choice), inputs=[binary])
                correction.change(fn=lambda correction: correction_feedback_handler(
                    correction), inputs=[correction])
                submit_button.click(fn=self.feedback_callback)

            return demo.launch(*args, **kwargs)



"""
Chainlink is a Python package that serves as an SDK to help developers and data scientists quickly build
standalone chat interfaces for LLMs. Our goal is to be able to provide an easy-to-use interface for
developers to deploy any LLM or chatbot application to an easy-to-use interface.
"""

from .model import Model, BaseModel
from .app import App, feedback

__version__ = '0.0.1'



class BaseModel:
    """
    Base model to extend when creating custom models for Chainlink

    To ensure interoperability with Chainlink applications, ensure that the `run` method is
    overwritten. The method should accept one single string parameter and return a string
    """

    def run(self):
        raise NotImplemented



from setuptools import setup
from llmlink import __version__

with open('requirements.txt', 'r') as f:
    requirements = [line for line in f.read().splitlines() if line]

setup(
    name='llmlink',
    version=__version__,
    packages=[
        'llmlink',
        'llmlink.app',
        'llmlink.model'
    ],
    description='LLMLink is a Python package that allows users to easily create LLM-powered Gradio applications',
    long_description=open('README.md', 'r').read(),
    long_description_content_type = 'text/markdown',
    author='Jacob Renn',
    author_email='jacob.renn@squared.ai',
    url='https://github.com/jacobrenn/chainlink.git',
    license='Apache 2.0',
    license_files='LICENSE',
    install_requires=requirements
)



from langchain.agents import initialize_agent, ZeroShotAgent, AgentExecutor
from langchain import OpenAI, LLMChain, HuggingFacePipeline, PromptTemplate
from langchain.memory import ConversationBufferMemory
from langchain.chat_models import ChatOpenAI
from transformers import Pipeline
from .BaseModel import BaseModel


class Model(BaseModel):
    """
    A Model class to be used with applications in Chainlink

    Parameters
    ----------
    model : str or huggingface pipeline (default 'gpt-4')
        The name of the model or the pipeline itself to be used
    model_type : str (default 'chat_openai')
        The type of model to be used. Currently supports 'chat_openai', 'openai', or 'huggingface'
    memory : bool (default False)
        Whether to support memory when chatting with the model
    tools : list or None (default None)
        List of tools to give the model access to, thus creating an agent
    temperature : int (default 0)
        Temperature value to use with OpenAI models
    huggingface_task : str or None (default None)
        The task to use for the huggingface pipeline
    huggingface_model_kwargs : dict or None (default None)
        Any keyword arguments to use to load the huggingface model
    huggingface_pipeline_kwargs : dict or None (default None)
        Any keyword arguments to use to load the huggingface pipeline
    openai_api_key : str or None (default None)
        Your OpenAI API key
    """

    def __init__(
            self,
            model='gpt-4',
            model_type='chat_openai',
            memory=False,
            tools=None,
            temperature=0,
            huggingface_task=None,
            huggingface_model_kwargs=None,
            huggingface_pipeline_kwargs=None,
            openai_api_key=None
    ):
        self.model = model
        self.model_type = model_type
        self.memory = memory
        self.tools = tools
        self.temperature = temperature
        self.huggingface_task = huggingface_task
        self.huggingface_model_kwargs = huggingface_model_kwargs
        self.huggingface_pipeline_kwargs = huggingface_pipeline_kwargs
        self.openai_api_key = openai_api_key

        self._agent = self._initialize_model()

    def _initialize_model(self):
        """
        Initialize the model

        NOTE: THIS FUNCTION IS NOT INTENDED TO BE CALLED BY THE USER
        """
        if self.model_type == 'openai':
            llm = OpenAI(
                model=self.model,
                openai_api_key=self.openai_api_key,
                temperature=self.temperature
            )
        elif self.model_type == 'chat_openai':
            llm = ChatOpenAI(
                model=self.model,
                openai_api_key=self.openai_api_key,
                temperature=self.temperature
            )
        elif self.model_type == 'huggingface':
            if isinstance(self.model, Pipeline):
                llm = HuggingFacePipeline(
                    pipeline=self.model
                )
            else:
                llm = HuggingFacePipeline.from_model_id(
                    model_id=self.model,
                    task=self.huggingface_task,
                    model_kwargs=self.huggingface_model_kwargs,
                    pipeline_kwargs=self.huggingface_pipeline_kwargs
                )

        if self.tools:
            if self.memory:
                prefix = """Have a conversation with a human, answering the following questions as best you can. You have access to the following tools:"""
                suffix = """Begin!"

                {chat_history}
                Question: {input}
                {agent_scratchpad}"""

                prompt = ZeroShotAgent.create_prompt(
                    self.tools,
                    prefix=prefix,
                    suffix=suffix,
                    input_variables=[
                        "input", "chat_history", "agent_scratchpad"],
                )
                memory = ConversationBufferMemory(memory_key="chat_history")
                llm_chain = LLMChain(llm=llm, prompt=prompt)
                agent = ZeroShotAgent(
                    llm_chain=llm_chain, tools=self.tools, verbose=True)
                return AgentExecutor.from_agent_and_tools(
                    agent=agent, tools=self.tools, verbose=True, memory=memory
                )
            else:
                return initialize_agent(
                    tools=self.tools,
                    llm=llm,
                    verbose=True
                )

        else:
            if self.memory:
                prompt = PromptTemplate(
                    template="""Have a conversation with a human, answering the following questions as best you can.
                    Begin!

                    {chat_history}
                    Question: {input}
                    """,
                    input_variables=['input', 'chat_history']
                )
                memory = ConversationBufferMemory(memory_key='chat_history')
                return LLMChain(llm=llm, prompt=prompt, memory=memory)
            else:
                return LLMChain(llm=llm, prompt=PromptTemplate(template='{input}', input_variables=['input']))

    def run(self, text):
        """
        Run the model on some text

        Parameters
        ----------
        text : str
            The user prompt to run

        Returns
        -------
        response : str
            The response from the model
        """
        response = self._agent.run(text)
        if response.startswith('AI: '):
            response = response[4:]
        return response



import csv
import os

feedback = {
    'user_msg': None,
    'bot_msg': None,
    'binary_feedback': None,
    'correction': None
}


def binary_feedback_handler(choice):
    feedback['binary_feedback'] = 1 if choice == 'Good Response' else -1


def correction_feedback_handler(correction):
    feedback['correction'] = correction


def print_feedback():
    """
    Default function to submit feedback - simply prints the feedback to the console
    """
    if feedback['user_msg'] is not None and feedback['bot_msg'] is not None:
        print(f'User Messsage: {feedback["user_msg"]}')
        print(f'Bot Message: {feedback["bot_msg"]}')
        print(f'Binary Feedback: {feedback["binary_feedback"]}')
        print(f'Correction: {feedback["correction"]}')

    feedback['user_msg'] = None
    feedback['bot_msg'] = None
    feedback['binary_feedback'] = None
    feedback['correction'] = None


def save_feedback_to_csv(path='./feedback.csv'):
    if feedback['user_msg'] is not None and feedback['bot_msg'] is not None:
        csv_exists = os.path.isfile(path)
        with open(path, 'a', newline='') as f:
            writer = csv.writer(f)
            if not csv_exists:
                writer.writerow(['user_prompt', 'bot_response',
                                'binary_feedback', 'correction'])
            writer.writerow([feedback['user_msg'], feedback['bot_msg'],
                            feedback['binary_feedback'], feedback['correction']])

    feedback['user_msg'] = None
    feedback['bot_msg'] = None
    feedback['binary_feedback'] = None
    feedback['correction'] = None
